@@ -11,19 +11,21 @@ import urlparse
 import urllib2
 import urllib
 import ConfigParser
+from distutils.sysconfig import get_python_version
 from email.FeedParser import FeedParser
-from pip.locations import bin_py, site_packages
+from pip.locations import bin_py
 from pip.exceptions import InstallationError, UninstallationError
 from pip.vcs import vcs
 from pip.log import logger
 from pip.util import display_path, rmtree, format_size
 from pip.util import splitext, ask, backup_dir
-from pip.util import url_to_filename, filename_to_url
-from pip.util import is_url, is_filename
-from pip.util import renames, normalize_path, is_framework_layout
+from pip.util import url_to_path, path_to_url
+from pip.util import is_url, is_installable_dir, is_archive_file, is_local, dist_is_local
+from pip.util import renames, normalize_path, egg_link_path
 from pip.util import make_path_relative, is_svn_page, file_contents
 from pip.util import has_leading_dir, split_leading_dir
 from pip.util import get_file_content
+from pip.util import in_venv
 from pip import call_subprocess
 from pip.backwardcompat import any, md5
 from pip.index import Link
@@ -59,7 +61,7 @@ class InstallRequirement(object):
     def from_editable(cls, editable_req, comes_from=None, default_vcs=None):
         name, url = parse_editable(editable_req, default_vcs)
         if url.startswith('file:'):
-            source_dir = url_to_filename(url)
+            source_dir = url_to_path(url)
         else:
             source_dir = None
         return cls(name, comes_from, source_dir=source_dir, editable=True, url=url)
@@ -67,21 +69,30 @@ class InstallRequirement(object):
     @classmethod
     def from_line(cls, name, comes_from=None):
         """Creates an InstallRequirement from a name, which might be a
-        requirement, filename, or URL.
+        requirement, directory containing 'setup.py', filename, or URL.
         """
         url = None
         name = name.strip()
         req = name
+        path = os.path.normpath(os.path.abspath(name))
+
         if is_url(name):
             url = name
             ## FIXME: I think getting the requirement here is a bad idea:
             #req = get_requirement_from_url(url)
             req = None
-        elif is_filename(name):
-            if not os.path.exists(name):
+        elif os.path.isdir(path):
+            if not is_installable_dir(path):
+                raise InstallationError("Directory %r is not installable. File 'setup.py' not found."
+                                        % name)
+            url = path_to_url(name)
+            #req = get_requirement_from_url(url)
+            req = None
+        elif is_archive_file(path):
+            if not os.path.isfile(path):
                 logger.warn('Requirement %r looks like a filename, but the file does not exist'
                             % name)
-            url = filename_to_url(name)
+            url = path_to_url(name)
             #req = get_requirement_from_url(url)
             req = None
         return cls(req, comes_from, url=url)
@@ -377,20 +388,13 @@ execfile(__file__)
         if not self.check_if_exists():
             raise UninstallationError("Cannot uninstall requirement %s, not installed" % (self.name,))
         dist = self.satisfied_by or self.conflicts_with
-        paths_to_remove = UninstallPathSet(dist, sys.prefix)
+
+        paths_to_remove = UninstallPathSet(dist)
 
         pip_egg_info_path = os.path.join(dist.location,
                                          dist.egg_name()) + '.egg-info'
         easy_install_egg = dist.egg_name() + '.egg'
-        # This won't find a globally-installed develop egg if
-        # we're in a virtualenv.
-        # (There doesn't seem to be any metadata in the
-        # Distribution object for a develop egg that points back
-        # to its .egg-link and easy-install.pth files).  That's
-        # OK, because we restrict ourselves to making changes
-        # within sys.prefix anyway.
-        develop_egg_link = os.path.join(site_packages,
-                                        dist.project_name) + '.egg-link'
+        develop_egg_link = egg_link_path(dist)
         if os.path.exists(pip_egg_info_path):
             # package installed by pip
             paths_to_remove.add(pip_egg_info_path)
@@ -428,8 +432,6 @@ execfile(__file__)
             easy_install_pth = os.path.join(os.path.dirname(develop_egg_link),
                                             'easy-install.pth')
             paths_to_remove.add_pth(easy_install_pth, dist.location)
-            # fix location (so we can uninstall links to sources outside venv)
-            paths_to_remove.location = normalize_path(develop_egg_link)
 
         # find distutils scripts= scripts
         if dist.has_metadata('scripts') and dist.metadata_isdir('scripts'):
@@ -519,17 +521,21 @@ execfile(__file__)
             return
         temp_location = tempfile.mkdtemp('-record', 'pip-')
         record_filename = os.path.join(temp_location, 'install-record.txt')
-        ## FIXME: I'm not sure if this is a reasonable location; probably not
-        ## but we can't put it in the default location, as that is a virtualenv symlink that isn't writable
-        header_dir = os.path.join(os.path.dirname(os.path.dirname(self.source_dir)), 'lib', 'include')
+
+        install_args = [sys.executable, '-c',
+                        "import setuptools; __file__=%r; execfile(%r)" % (self.setup_py, self.setup_py),
+                        'install', '--single-version-externally-managed', '--record', record_filename]
+
+        if in_venv():
+            ## FIXME: I'm not sure if this is a reasonable location; probably not
+            ## but we can't put it in the default location, as that is a virtualenv symlink that isn't writable
+            install_args += ['--install-headers',
+                             os.path.join(sys.prefix, 'include', 'site',
+                                          'python' + get_python_version())]
         logger.notify('Running setup.py install for %s' % self.name)
         logger.indent += 2
         try:
-            call_subprocess(
-                [sys.executable, '-c',
-                 "import setuptools; __file__=%r; execfile(%r)" % (self.setup_py, self.setup_py),
-                 'install', '--single-version-externally-managed', '--record', record_filename,
-                 '--install-headers', header_dir] + install_options,
+            call_subprocess(install_args + install_options,
                 cwd=self.source_dir, filter_stdout=self._filter_install, show_stdout=False)
         finally:
             logger.indent -= 2
@@ -763,6 +769,43 @@ class RequirementSet(object):
             req.uninstall(auto_confirm=auto_confirm)
             req.commit_uninstall()
 
+    def locate_files(self):
+        ## FIXME: duplicates code from install_files; relevant code should
+        ##        probably be factored out into a separate method
+        unnamed = list(self.unnamed_requirements)
+        reqs = self.requirements.values()
+        while reqs or unnamed:
+            if unnamed:
+                req_to_install = unnamed.pop(0)
+            else:
+                req_to_install = reqs.pop(0)
+            install_needed = True
+            if not self.ignore_installed and not req_to_install.editable:
+                req_to_install.check_if_exists()
+                if req_to_install.satisfied_by:
+                    if self.upgrade:
+                        req_to_install.conflicts_with = req_to_install.satisfied_by
+                        req_to_install.satisfied_by = None
+                    else:
+                        install_needed = False
+                if req_to_install.satisfied_by:
+                    logger.notify('Requirement already satisfied '
+                                  '(use --upgrade to upgrade): %s'
+                                  % req_to_install)
+
+            if req_to_install.editable:
+                if req_to_install.source_dir is None:
+                    req_to_install.source_dir = req_to_install.build_location(self.src_dir)
+            elif install_needed:
+                req_to_install.source_dir = req_to_install.build_location(self.build_dir, not self.is_download)
+
+            if req_to_install.source_dir is not None and not os.path.isdir(req_to_install.source_dir):
+                raise InstallationError('Could not install requirement %s '
+                                       'because source folder %s does not exist '
+                                       '(perhaps --no-download was used without first running '
+                                       'an equivalent install with --no-install?)'
+                                       % (req_to_install, req_to_install.source_dir))
+
     def install_files(self, finder, force_root_egg_info=False, bundle=False):
         unnamed = list(self.unnamed_requirements)
         reqs = self.requirements.values()
@@ -788,7 +831,7 @@ class RequirementSet(object):
                 logger.notify('Obtaining %s' % req_to_install)
             elif install:
                 if req_to_install.url and req_to_install.url.lower().startswith('file:'):
-                    logger.notify('Unpacking %s' % display_path(url_to_filename(req_to_install.url)))
+                    logger.notify('Unpacking %s' % display_path(url_to_path(req_to_install.url)))
                 else:
                     logger.notify('Downloading/unpacking %s' % req_to_install)
             logger.indent += 2
@@ -913,14 +956,21 @@ class RequirementSet(object):
                 return
         temp_dir = tempfile.mkdtemp()
         if link.url.lower().startswith('file:'):
-            source = url_to_filename(link.url)
+            source = url_to_path(link.url)
             content_type = mimetypes.guess_type(source)[0]
-            self.unpack_file(source, location, content_type, link)
+            if os.path.isdir(source):
+                # delete the location since shutil will create it again :(
+                if os.path.isdir(location):
+                    shutil.rmtree(location)
+                shutil.copytree(source, location)
+            else:
+                self.unpack_file(source, location, content_type, link)
             return
         md5_hash = link.md5_hash
         target_url = link.url.split('#', 1)[0]
         target_file = None
         if self.download_cache:
+            self.download_cache = os.path.expanduser(self.download_cache)
             if not os.path.isdir(self.download_cache):
                 logger.indent -= 2
                 logger.notify('Creating supposed download cache at %s' % self.download_cache)
@@ -1295,23 +1345,23 @@ def parse_requirements(filename, finder=None, comes_from=None, options=None):
             # No longer used, but previously these were used in
             # requirement files, so we'll ignore.
             pass
-        elif finder and line.startswith('-f') or line.startswith('--find-links'):
+        elif line.startswith('-f') or line.startswith('--find-links'):
             if line.startswith('-f'):
                 line = line[2:].strip()
             else:
                 line = line[len('--find-links'):].strip().lstrip('=')
             ## FIXME: it would be nice to keep track of the source of
             ## the find_links:
-            finder.find_links.append(line)
+            if finder: finder.find_links.append(line)
         elif line.startswith('-i') or line.startswith('--index-url'):
             if line.startswith('-i'):
                 line = line[2:].strip()
             else:
                 line = line[len('--index-url'):].strip().lstrip('=')
-            finder.index_urls = [line]
+            if finder: finder.index_urls = [line]
         elif line.startswith('--extra-index-url'):
             line = line[len('--extra-index-url'):].strip().lstrip('=')
-            finder.index_urls.append(line)
+            if finder: finder.index_urls.append(line)
         else:
             comes_from = '-r %s (line %s)' % (filename, line_number)
             if line.startswith('-e') or line.startswith('--editable'):
@@ -1331,7 +1381,7 @@ def parse_editable(editable_req, default_vcs=None):
     url = editable_req
     if os.path.isdir(url) and os.path.exists(os.path.join(url, 'setup.py')):
         # Treating it as code that has already been checked out
-        url = filename_to_url(url)
+        url = path_to_url(url)
     if url.lower().startswith('file:'):
         return None, url
     for version_control in vcs:
@@ -1370,35 +1420,26 @@ def parse_editable(editable_req, default_vcs=None):
 class UninstallPathSet(object):
     """A set of file paths to be removed in the uninstallation of a
     requirement."""
-    def __init__(self, dist, restrict_to_prefix):
+    def __init__(self, dist):
         self.paths = set()
         self._refuse = set()
         self.pth = {}
-        self.prefix = normalize_path(restrict_to_prefix)
         self.dist = dist
-        self.location = normalize_path(dist.location)
         self.save_dir = None
         self._moved_paths = []
 
     def _permitted(self, path):
         """
-        Return True if the given path is one we are permitted to remove,
-        False otherwise.
+        Return True if the given path is one we are permitted to
+        remove/modify, False otherwise.
 
         """
-        ok_prefixes = [self.prefix]
-        # Yep, we are special casing the framework layout of MacPython here
-        if is_framework_layout(sys.prefix):
-            for location in ('/Library', '/usr/local'):
-                if path.startswith(location):
-                    ok_prefixes.append(location)
-        return any([path.startswith(prefix) for prefix in ok_prefixes])
+        return is_local(path)
         
     def _can_uninstall(self):
-        if not self._permitted(self.location):
+        if not dist_is_local(self.dist):
             logger.notify("Not uninstalling %s at %s, outside environment %s"
-                          % (self.dist.project_name, self.location,
-                             self.prefix))
+                          % (self.dist.project_name, normalize_path(self.dist.location), sys.prefix))
             return False
         return True
 
@@ -1427,7 +1468,7 @@ class UninstallPathSet(object):
         /a/path/to/a/file.txt are both in the set, leave only the
         shorter path."""
         short_paths = set()
-        for path in sorted(paths, lambda x, y: cmp(len(x), len(y))):
+        for path in sorted(paths, key=len):
             if not any([(path.startswith(shortpath) and
                          path[len(shortpath.rstrip(os.path.sep))] == os.path.sep)
                         for shortpath in short_paths]):
@@ -1457,8 +1498,8 @@ class UninstallPathSet(object):
                 self.save_dir = tempfile.mkdtemp(suffix='-uninstall',
                                                  prefix='pip-')
                 for path in paths:
-                    new_path = os.path.join(self.save_dir,
-                                            path.lstrip(os.path.sep))
+                    new_path = os.path.splitdrive(path)[1].lstrip(os.path.sep)
+                    new_path = os.path.join(self.save_dir, new_path)
                     logger.info('Removing file or directory %s' % path)
                     self._moved_paths.append(path)
                     renames(path, new_path)
